



##### Two threads -> one for receiving jobs and responding about the status, second for execution


import queue
import threading
import time
import sys



threads=[]


class Thread(threading.Thread):
    def __init__(self,name,function,args,frequency=0.5,daemon=True):
        super().__init__(daemon=daemon)
        self.name=name
        self.exit_flag=False
        self.function=function
        self.args=args
        self.frequency=frequency
        
    def run(self):
        print(f"Starting: {self.name}")

        
        while not self.exit_flag:
            self.function(self.args)
            
            time.sleep(self.frequency)
        print(f"Exiting: {self.name}")
        
    def run_function(self,function):
        function()
        
    
    

from distjob_context import djc
class DistJobServer:
    
    def __init__(self,host="0.0.0.0",port=10100):
        self.host=host
        self.port=port
        
        
        djc.host=host
        djc.port=port

        
        global threads
        
           
        
        print("Main Thread")
        
        lock=threading.Lock()
        queue1=queue.Queue()
        
        def run_api(args):
            from distjob_api import run_api
            run_api()
            print("API RUNNING!")
            
        
        def new_thread(name,function,function_args,frequency=0.5):
            global threads
            thread=Thread(name,function,function_args,frequency)
            threads.append(thread)
            thread.start()
        
        
        def run_worker(args): #needs to have one parameter
            from distjob_worker import run
            try:
                run()
            except (KeyboardInterrupt, SystemExit): #close thread on CTRL+C
                djc.ARE_ALL_THREADS_FINISHED=True
                sys.exit()
       
        
    
        new_thread("ApplicationThread-1",run_worker,1)
        
            
        print("LOADING DISTJOB API...")
        new_thread("APIThread-4",run_api,None,10)
        time.sleep(1)
        
        
        
        
        def check_for_finishing_of_threads(args):
            print("CHECK",args,djc.ARE_ALL_THREADS_FINISHED)
            if djc.ARE_ALL_THREADS_FINISHED:
                for thread in threads:
                    thread.exit_flag=True
                    thread.join()
                
        
        new_thread("ThreadingMonitoringThread-6",check_for_finishing_of_threads,None,3)
        time.sleep(1)
        
    
        time.sleep(1)
         
        
        
        for thread in threads:
            thread.exit_flag=True
            thread.join()
        
        
        print("Exiting Main Thread")

        

if __name__=="__main__":
    
    try:
        dist_job_server=DistJobServer(port=djc.port)
    except (KeyboardInterrupt, SystemExit): #close thread on CTRL+C
        djc.ARE_ALL_THREADS_FINISHED=True
        sys.exit()


"""
This file assigns jobs to available machines based on triggers
"""

import time

import sys
from distjob_context import djc

import datetime

print(datetime.datetime.now().timestamp()>datetime.datetime.now().timestamp())


def run():
    
    done=False
    
    i=0
    while not done:
        
        
        try:
            print("JOBS",djc.port,djc.jobs)
            print(i)
            
            
            ready_jobs=[x for x in djc.jobs if x.start_datetime.timestamp()<=datetime.datetime.now().timestamp() and not x.is_assigned and not x.is_done]
            if len(ready_jobs)>0:
                machine=djc.assign_job_to_idle_machine(ready_jobs[0])
                print(machine,ready_jobs[0])
            
            
            
            
            
            assigned_jobs=[x for x in djc.jobs if x.is_assigned and not x.is_done]
            for i,job in enumerate(assigned_jobs):
                print("CHECKING JOB",job.uid)
                if job.url=="test":
                    djc.process_test_job(job)
                    
                
                
                matching_machine=[x for x in djc.machines if x.processing_job==job][0]
            
                is_job_done=djc.check_if_job_done(matching_machine,job)
                
                    
            
            
                
            
            
            
            time.sleep(0.5)
            i+=1
            
            #if i>100:
                #done=True
                
        except (KeyboardInterrupt, SystemExit): #close thread on CTRL+C
            djc.ARE_ALL_THREADS_FINISHED=True
            
            sys.exit()


# -*- coding: utf-8 -*-
"""
Created on Wed Feb  8 01:33:47 2023

@author: EUROCOM
"""
import requests
import json
import datetime
#'2023-01-01T12:00:00.000Z'
data="{'url': 'test', 'data': 'string', 'priority': 0, 'start_datetime': null, 'check_condition_request_url': '/api/v1/test_condition', 'method': 'POST'}"

data={'url': 'test', 'data': {}, 'priority': 0, 'start_datetime': '2023-01-01T12:00:00.000Z', 'check_condition_request_url': '/api/v1/test_condition', 'method': 'POST'}

requests.post(url="http://127.0.0.1:10102/api/v1/jobs",data=json.dumps(data))


import requests
import json
import datetime 

class Job:
    instance_counter=0
    def __init__(self, url, data, priority, start_datetime, check_condition_request_url, method="POST"):
        self.url=url
        self.data=data
        self.priority=priority
        self.start_datetime=start_datetime
        self.check_condition_request_url=check_condition_request_url #should return True or False if job is done
        self.method=method
        self.is_assigned=False
        self.is_done=False
        
        self.__class__.instance_counter += 1
        self.uid = self.instance_counter
        


class Machine:
    instance_counter=0
    def __init__(self, host, port):
        self.host=host
        self.port=port
        self.processing_job=None
        
        self.__class__.instance_counter += 1
        self.uid = self.instance_counter
        
        
        


    


class DistJobContext:
    def __init__(self,host="0.0.0.0",port=10100):
        self.ARE_ALL_THREADS_FINISHED=False
        
        self.host=host
        self.port=port        
        
        
        self.jobs=[]
        self.machines=[]
        self.triggers=[]
        self.test_condition=False
        self.test_jobs_progress_dict={}
        
        #self.create_test_job()
        #self.create_test_job()
        #self.create_test_job()
        
    def create_test_job(self):
        data={"url": "test","data": {},"priority": 0,"start_datetime": "2023-01-01T12:00:00.000Z","check_condition_request_url": "/api/v1/test_condition","method": "POST"}
        self.jobs.append(Job("/api/v1/jobs",data,0,datetime.datetime.now(),"/api/v1/test_condition"))
        
    
    
    
        
    def process_test_job(self,job):
        print("TEST_PROGRESS",self.test_jobs_progress_dict)
        if job.uid in self.test_jobs_progress_dict:
            if self.test_jobs_progress_dict[job.uid]<10:
                
                self.test_jobs_progress_dict[job.uid]+=1
                self.test_condition=False
            else:
                job.is_done=True
                job.is_assigned=False  
                self.test_condition=True #len([x for x in self.jobs if x.is_done])==len(self.jobs) #True if all jobs done
                print("TEST CONDITION SET",self.test_condition)
        else:
            self.test_jobs_progress_dict[job.uid]=0
                     
            
        
    
        
        
        
    def assign_job_to_idle_machine(self,job):
        
        for i,machine in enumerate(self.machines):
            if machine.processing_job is None:
                if job.url=="test":
                    self.machines[i].processing_job=job
                    job.is_assigned=True
                    print("Machine",i,"was assigned")
                    return(self.machines[i])
                
                response=None
                url="http://"+machine.host+":"+str(machine.port)+job.url
                #job.data should be in dict format (json)
                print("URL",job.url)
                print("DATA",job.data,type(job.data))
                try:
                    if job.method=="POST":
                        response=requests.post(url,json.dumps(job.data))
                        result=json.loads(response.content)
                    elif job.method=="GET":
                        response=requests.post(job.url,json.dumps(job.data))
                        result=json.loads(response.content)
                    elif job.method=="DELETE":
                        response=requests.delete(job.url,json.dumps(job.data))
                        result=json.loads(response.content)
                    elif job.method=="PUT":
                        response=requests.put(job.url,json.dumps(job.data))
                        result=json.loads(response.content)
                
                
                    print("JOB ASSIGNMENT RESPONSE:",response)
                    if response is not None:
                        if response.status_code==200:
                            
                            self.machines[i].processing_job=job
                            job.is_assigned=True
                            print("Machine",i,"was assigned")
                            return(self.machines[i])
                        else:
                            print("Machine couldnt be assigned, wrong status code",response.status_code)
                            return(None)
                    
                except requests.exceptions.ConnectionError:
                    print("Connection error")
                
        print("No machine could be assigned")
        return(None)
    
    def unassign_job_from_machine(self,job):
        
        for i,machine in enumerate(self.machines):
            if machine.processing_job==job:
                self.machines[i].processing_job=None
                job.is_assigned=False
                print("Machine",machine.uid,"was unassigned")
                return(self.machines[i])
        print("No machine could be unassigned")
        return(None)
    
    
    
    def check_if_job_done(self, machine, job):
        
        is_running_locally_on_this_port=False
        if self.host.strip() in ["0.0.0.0","127.0.0.1"] and machine.host.strip() in ["0.0.0.0","127.0.0.1"]:
            is_running_locally_on_this_port=True
        if machine.host==self.host:
            is_running_locally_on_this_port=True
        
        
        if is_running_locally_on_this_port and machine.port==self.port: #if executed on this machine (e.g. test task), dont check test condition
            return(False)
        else:   
            url="http://"+machine.host+":"+str(machine.port)+job.check_condition_request_url
            print(url)
            response=requests.get(url=url)
            result=json.loads(response.content)
            print(result)
            is_job_done=result["condition"]
            print(job,"IS_JOB_DONE",is_job_done)
            
            if is_job_done:
                index=self.machines.index(machine)
                self.machines[index].processing_job=None
                job.is_assigned=False
                job.is_done=True
            
            return(result["condition"])
        
        
        
        
    
    ###### Jobs #####

    def get_job_by_uid(self,job_uid):
        matching_jobs=[job for job in self.jobs if job.uid==job_uid]
        assert len(matching_jobs)<=1
        if len(matching_jobs)==1:
            return(matching_jobs[0])
        elif len(matching_jobs)==0:
            print("Job was not found, couldn't be deleted")
            return(None)
        else:
            print("Multiple jobs with duplicate uid exist")
            return(None)



    def new_job(self, url, data, priority, start_datetime, check_condition_request_url):
        job=Job(url, data, priority, start_datetime, check_condition_request_url)
        
        self.jobs.append(job)
        return(job)
        
    
    def delete_job(self,job):
        index=self.jobs.index(job)
        self.jobs.pop(index)
        
     
        

    def update_job(self,job,url=None, data=None, priority=None, start_datetime=None, check_condition_request_url=None):
        index=self.jobs.index(job)
        if url is not None:
            self.jobs[index].url=url
        if data is not None:
            self.jobs[index].data=data
        if priority is not None:
            self.jobs[index].priority=priority
        if start_datetime is not None:
            self.jobs[index].start_datetime=start_datetime
        if check_condition_request_url is not None:
            self.jobs[index].check_condition_request_url=check_condition_request_url
            
       
        
    def update_job_by_uid(self,job_uid, url=None, data=None, priority=None, start_datetime=None, check_condition_request_url=None):
        job=self.get_job_by_uid(job_uid)
        try:
            self.update_job(job,url, data, priority, start_datetime, check_condition_request_url)
        except Exception as e:
            print(f"Job couldn't be updated: {e}")
        
    def delete_job_by_uid(self,job_uid):
        job=self.get_job_by_uid(job_uid)
        try:
            self.delete_job(job)
        except Exception as e:
            print(f"Job couldn't be deleted: {e}")
    
    
    
       
    
    ###### Machines #####

    def get_machine_by_uid(self,machine_uid):
        matching_machines=[machine for machine in self.machines if machine.uid==machine_uid]
        assert len(matching_machines)<=1
        if len(matching_machines)==1:
            return(matching_machines[0])
        elif len(matching_machines)==0:
            print("machine was not found, couldn't be deleted")
            return(None)
        else:
            print("Multiple machines with duplicate uid exist")
            return(None)



    def new_machine(self, host, port):
        machine=Machine(host, port)
        
        self.machines.append(machine)
        return(machine)
        
    
    def delete_machine(self,machine):
        index=self.machines.index(machine)
        self.machines.pop(index)
        
     
        

    def update_machine(self,machine,host=None, port=None):
        index=self.machines.index(machine)
        if host is not None:
            self.machines[index].host=host
        if port is not None:
            self.machines[index].port=port
       
        
    def update_machine_by_uid(self,machine_uid, host=None, port=None):
        machine=self.get_machine_by_uid(machine_uid)
        try:
            self.update_machine(machine, host, port)
        except Exception as e:
            print(f"machine couldn't be updated: {e}")
        
    def delete_machine_by_uid(self,machine_uid):
        machine=self.get_machine_by_uid(machine_uid)
        try:
            self.delete_machine(machine)
        except Exception as e:
            print(f"machine couldn't be deleted: {e}")
    
    
 
djc=DistJobContext()



from typing import Optional, Any, List, Dict, Union

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from pydantic import BaseModel


from distjob_context import djc


import uvicorn

import datetime

description = """
This API helps you to distribute jobs among many machines via API requests. 🚀
"""


tags_metadata=[
    {"name": "Get Methods", "description": ""},
    {"name": "Post Methods", "description": ""},
    {"name": "Delete Methods", "description": ""},
    {"name": "Put Methods", "description": ""},
]

app = FastAPI(title="Distjob API",
    description=description,
    version="1.0.0",
    openapi_tags=tags_metadata
    )

origins = ["*"]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
import json

class APIJob(BaseModel):
    url:Any="/api/v1/jobs"
    data:Any='{"url": "test","data": {},"priority": 0,"start_datetime": "2023-01-01T12:00:00.000Z","check_condition_request_url": "/api/v1/test_condition","method": "POST"}'
    priority:int
    start_datetime:datetime.datetime
    check_condition_request_url:str="/api/v1/test_condition"
    method:str="POST"
    
    
class APIMachine(BaseModel):
    host:str="127.0.0.1"
    port:int=10100  
    
    
    
    
    
    
    
    
    

class APINode(BaseModel):
    pos:Union[None,List[int]]=[0,0]
    typ:Union[None,str]="Custom"
    params:Union[None,Dict[str, Any]]
    fields:Optional[List[Any]]=[]
    is_active:bool=False
    pipeline_uid:int=0
    project_key:str=""
    
    
class APIEdge(BaseModel):
    from_node_uid:int
    to_node_uid:int
    channel:Any
    pipeline_uid:int=0
    project_key:str=""
    

class APIVariable(BaseModel):
    name:str=""
    value:str=""
    type:Optional[str]=""
    size:Optional[int]=0
    pipeline_uid:int=0
    project_key:str=""
    

class DeleteObject(BaseModel):
    project_key:str=""
        
class APIPipeline(BaseModel):
    name:str="untitled"
    start_node_uid:int=0
    is_active:bool=False
    #nodes_uids:List[int]
    #edges_uids:List[int]
    #variables_uids:List[int]
    active_nodes_uids:List[int]=[]
    remaining_nodes_uids:List[int]=[]
    workspace_uid:int=0
    project_key:str=""
    
    

class APIPopup(BaseModel):
    pos:List[int]=[0,0]
    typ:str="Custom"
    params:Dict[str, Any]
    project_key:str=""


class APIWorkspace(BaseModel):
    workspace_name: str=""
    project_key: str=""
 
import datetime
    
class APITrigger(BaseModel):
    trigger_name: str=""
    machine_uid: int=""
    pipeline_uid: int=""
    first_run_datetime: datetime.datetime
    frequency: int=""
    workspace_uid: int=0
    
class APIDatabase(BaseModel):
    database_name: str=""
    server: str
    port: int
    database: str
    username: str
    password: str    
    dialect: str
    workspace_uid: int=0
    
    
class APIDataset(BaseModel):
    dataset_name: str=""
    data: Any
    workspace_uid: int=0
    
    
class APIFile(BaseModel):
    file_name: str=""
    suffix: str=""
    data: Any
    workspace_uid: int=0
    
    
class APIScript(BaseModel):
    script_name: str=""
    text: str=""
    workspace_uid: int=0
    




@app.get("/api/v1/jobs")
def get_jobs():
    """
    Returns all jobs in a given server
    """
    return {"jobs": djc.jobs}


@app.get("/api/v1/job/{uid}")
def get_job(uid: Optional[int]=None):
    """
    uid: unique id of a job is required
    """
    matching_jobs=[x for x in djc.jobs if x.uid==uid]
    if len(matching_jobs)==1:
        job=matching_jobs[0]
        return {
            "uid":job.uid,
            "url":job.url,
            "data":job.data,
            "priority":job.priority,
            "start_datetime":job.start_datetime,
            "check_condition_request_url":job.check_condition_request_url,
            "method":job.method
            }
    else:
        return {"ok":False}
        
  
@app.post("/api/v1/jobs")
def new_job(job : APIJob):
    job=djc.new_job(job.url,job.data,job.priority,job.start_datetime,job.check_condition_request_url)
    return {
        "uid":job.uid,
        "url":job.url,
        "data":job.data,
        "priority":job.priority,
        "start_datetime":job.start_datetime,
        "check_condition_request_url":job.check_condition_request_url,
        "method":job.method
        }

@app.delete("/api/v1/job/{uid}")
def delete_job(uid: Optional[int]=None):  
    djc.delete_job_by_uid(uid)  
    return {"ok":True}

@app.delete("/api/v1/jobs")
def delete_jobs():  
    djc.jobs=[]
    return {"ok":True}

@app.put("/api/v1/job/{uid}")
def update_job(uid: Optional[int]=None, job:APIJob=None):
    djc.update_job_by_uid(uid, job.url,job.data,job.priority,job.start_datetime,job.check_condition_request_url)  
    return {"ok":True}



@app.get("/api/v1/machines")
def get_machines():
    """
    Returns all machines in a given server
    """
    return {"machines": djc.machines}


@app.get("/api/v1/machine/{uid}")
def get_machine(uid: Optional[int]=None):
    """
    uid: unique id of a machine is required
    """
    matching_machines=[x for x in djc.machines if x.uid==uid]
    if len(matching_machines)==1:
        machine=matching_machines[0]
        return {
            "uid":machine.uid,
            "host":machine.host,
            "port":machine.port,
            }
    else:
        return {"ok":False}
        
  
@app.post("/api/v1/machines")
def new_machine(machine : APIMachine):
    machine=djc.new_machine(machine.host,machine.port)
    return {
        "uid":machine.uid,
        "host":machine.host,
        "port":machine.port,
        }

@app.delete("/api/v1/machine/{uid}")
def delete_machine(uid: Optional[int]=None):  
    djc.delete_machine_by_uid(uid)  
    return {"ok":True}

@app.delete("/api/v1/machines")
def delete_machines():  
    djc.machines=[]
    return {"ok":True}

@app.put("/api/v1/machine/{uid}")
def update_machine(uid: Optional[int]=None, machine:APIMachine=None):
    djc.update_machine_by_uid(uid, machine.host,machine.port)  
    return {"ok":True}


@app.get("/api/v1/test_condition")
def test_condition():
    """
    Returns sample test condition
    """
    print("TEST_CONDITION",djc.test_condition)
    if djc.test_condition: #return True just once - turn off on get request
        djc.test_condition=False
        return {"condition": True}
    else:
        return {"condition": False}




def run_api():
    uvicorn.run(app, host=djc.host, port=djc.port)

if __name__=="__main__":
    run_api()